/* This file is part of the KDE project
   Copyright (C) 2000 Eva Brucherseifer <eva@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "kvcard_part.h"
#include "kvcardlayout.h"
#include "vcard.h"
#include "kvcardeditor.h"
#include "kvcardconfig.h"

#include <kinstance.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kfiledialog.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kaction.h>

#include <qfile.h>
#include <qtextstream.h>
#include <ktextbrowser.h>
#include <qvbox.h>
#include <qlayout.h>

KvcardPart::KvcardPart( QWidget *parentWidget, const char *widgetName,
                        QObject *parent, const char *name )
		: KParts::ReadWritePart(parent, name)
{
	// this is needed for tracking instantiation and destruction of parts.
        // the library containing the code for this part can be unloaded only
        // if no instances of the part are "alive" any more.
	setInstance( KvcardPartFactory::instance() );

	// default: empty data
	m_vcard = new VCard();

	// create our internal widget and its layout, notify the part
	m_view = new KTextBrowser(parentWidget, "KVCview");
	m_layout = new KvcardLayout(m_view);
	setWidget(m_view);

	// create our actions
	KStdAction::open(this, SLOT(fileOpen()), actionCollection());
	KStdAction::saveAs(this, SLOT(fileSaveAs()), actionCollection());
	KStdAction::save(this, SLOT(save()), actionCollection());
	KStdAction::preferences(this, SLOT(changePreferences()), actionCollection());
	(void)new KAction( "Kvcard &Editor", "edit", Qt::CTRL + Qt::Key_E, this, SLOT( editVCard() ), actionCollection(), "view_kvcard_editor" );

	// set our XML-UI resource file
	setXMLFile("kvcard_part.rc");

	// we are read-write by default and data is not modified yet
	setReadWrite(true);
	setModified(false);
}

KvcardPart::~KvcardPart()
{}

void KvcardPart::setReadWrite(bool rw)
{
	// get a handle on our Edit action and make sure it is valid
	KAction *edit = actionCollection()->action("view_kvcard_editor");
	if (!edit)
		return;

	// if so, we either enable or disable it based on the current state
	edit->setEnabled(rw);
	ReadWritePart::setReadWrite(rw);
}

void KvcardPart::setModified(bool modified)
{
	// get a handle on our Save action and make sure it is valid
	KAction *save = actionCollection()->action(KStdAction::stdName(KStdAction::Save));
	if (!save)
		return;

	// if so, we either enable or disable it based on the current state
	save->setEnabled(modified);

	// in any event, we want our parent to do it's thing
	ReadWritePart::setModified(modified);
}

bool KvcardPart::openFile()
{
	QFile file(m_file);
	if (file.open(IO_ReadOnly) == false)
		return false;

	// our example widget is text-based, so we use QTextStream instead
	// of a raw QDataStream
	QTextStream stream(&file);
	stream.setEncoding(QTextStream::Locale);
	QString filecontent = stream.read();

	int errornumber;
	int linenumber;
	if (m_vcard)
		delete m_vcard;
	m_vcard = VCard::parseVCard(filecontent, &errornumber, &linenumber);

	if (!m_vcard)
	{
		QString message = i18n("Couldn't open file\n") + m_file;
		message += i18n(" Error in line %1").arg(linenumber) + ": ";
                message += VCard::getError(errornumber);
		KMessageBox::error(widget(), message);
		return false;
	}

	// output
	m_layout->displayVCard(m_vcard);
	emit setStatusBarText( m_url.prettyURL() );

	return true;
}

bool KvcardPart::saveFile()
{
	// if we aren't read-write, return immediately
	if (isReadWrite() == false)
		return false;

	// m_file is always local, so we use QFile
	QFile file(m_file);
	if (file.open(IO_WriteOnly) == false)
		return false;

	// use QTextStream to dump the text to the file
	QTextStream stream(&file);
	stream << m_vcard->getVCard();

	file.close();
	return true;
}

void KvcardPart::fileOpen()
{
	// this slot is called whenever the File->Open menu is selected,
	// the Open shortcut is pressed (usually CTRL+O) or the Open toolbar
	// button is clicked
	QString file_name = KFileDialog::getOpenFileName();

        // if the user entered a valid URL, we open it. We use openURL()
        // here, not our own openFile(), because we want the framework
        // to fetch the file via e.g. a network connection if it's non-local.
        // The framework will call KPart::openFile() after fetching the data
        // into a local, temporary file.
	if (file_name.isEmpty() == false)
		openURL(file_name);
}

void KvcardPart::fileSaveAs()
{
	// this slot is called whenever the File->Save As menu is selected,
	QString file_name = KFileDialog::getSaveFileName();
	if (file_name.isEmpty() == false)
		saveAs(file_name);
}

void KvcardPart::changePreferences()
{
	KvcardConfig* pref = new KvcardConfig (m_view, "configdialog", true);
	pref->load();
	if (pref->exec()==QDialog::Accepted)
	{
		pref->save();
		m_layout->syncPreferences();
		m_layout->displayVCard(m_vcard);
	}
	delete pref;
}


void KvcardPart::editVCard()
{
	KvcardEditor* editor = new KvcardEditor(m_view, "editor", true);
	editor->setVCard(m_vcard);
	if (editor->exec()==QDialog::Accepted)
	{
		*m_vcard = *(editor->vcard());
		m_layout->displayVCard(m_vcard);
		this->setModified(true);
	}
}


// It's usually safe to leave the factory code alone.. with the
// notable exception of the KAboutData data
#include <kaboutdata.h>
#include <klocale.h>

KInstance*  KvcardPartFactory::s_instance = 0L;
KAboutData* KvcardPartFactory::s_about = 0L;

KvcardPartFactory::KvcardPartFactory()
		: KParts::Factory()
{}

KvcardPartFactory::~KvcardPartFactory()
{
	delete s_instance;
	delete s_about;

	s_instance = 0L;
}

KParts::Part* KvcardPartFactory::createPartObject( QWidget *parentWidget, const char *widgetName,
        QObject *parent, const char *name,
        const char *classname, const QStringList &args )
{
	// Create an instance of our Part
	KvcardPart* obj = new KvcardPart( parentWidget, widgetName, parent, name );

	// See if we are to be read-write or not
	if (QCString(classname) == "KParts::ReadOnlyPart")
		obj->setReadWrite(false);

	return obj;
}

KInstance* KvcardPartFactory::instance()
{
	if( !s_instance )
	{
		s_about = new KAboutData("kvcardpart", I18N_NOOP("KvcardPart"), "0.1");
		s_about->addAuthor("Eva Brucherseifer", 0, "eva@kde.org");
		s_instance = new KInstance(s_about);
	}
	return s_instance;
}

extern "C"
{
	void* init_libkvcardpart()
	{
		return new KvcardPartFactory;
	}
};

#include "kvcard_part.moc"
